@@ -23,8 +23,8 @@ import java.io.ByteArrayOutputStream;
 import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
-import java.io.OutputStream;
 import java.util.Enumeration;
+import java.util.HashMap;
 import java.util.Map;
 import java.util.Properties;
 import java.util.zip.ZipEntry;
@@ -32,10 +32,10 @@ import org.apache.felix.framework.Logger;
 import org.apache.felix.framework.util.FelixConstants;
 import org.apache.felix.framework.util.JarFileX;
 import org.apache.felix.framework.util.Util;
-import org.apache.felix.moduleloader.IContent;
+import org.apache.felix.framework.resolver.Content;
 import org.osgi.framework.Constants;
 
-public class JarContent implements IContent
+public class JarContent implements Content
 {
     private static final int BUFSIZE = 4096;
     private static final transient String EMBEDDED_DIRECTORY = "-embedded";
@@ -46,16 +46,20 @@ public class JarContent implements IContent
     private final Object m_revisionLock;
     private final File m_rootDir;
     private final File m_file;
-    private JarFileX m_jarFile = null;
-    private int m_libCount = 0;
+    private final JarFileX m_jarFile;
+    private final boolean m_isJarFileOwner;
+    private Map m_nativeLibMap;
 
-    public JarContent(Logger logger, Map configMap, Object revisionLock, File rootDir, File file)
+    public JarContent(Logger logger, Map configMap, Object revisionLock, File rootDir,
+        File file, JarFileX jarFile)
     {
         m_logger = logger;
         m_configMap = configMap;
         m_revisionLock = revisionLock;
         m_rootDir = rootDir;
         m_file = file;
+        m_jarFile = (jarFile == null) ? openJarFile(m_file) : jarFile;
+        m_isJarFileOwner = (jarFile == null);
     }
 
     protected void finalize()
@@ -63,11 +67,11 @@ public class JarContent implements IContent
         close();
     }
 
-    public synchronized void close()
+    public void close()
     {
         try
         {
-            if (m_jarFile != null)
+            if (m_isJarFileOwner)
             {
                 m_jarFile.close();
             }
@@ -76,20 +80,12 @@ public class JarContent implements IContent
         {
             m_logger.log(
                 Logger.LOG_ERROR,
-                "JarContent: Unable to open JAR file.", ex);
+                "JarContent: Unable to close JAR file.", ex);
         }
-
-        m_jarFile = null;
     }
 
-    public synchronized boolean hasEntry(String name) throws IllegalStateException
+    public boolean hasEntry(String name) throws IllegalStateException
     {
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-        }
-
         try
         {
             ZipEntry ze = m_jarFile.getEntry(name);
@@ -104,14 +100,8 @@ public class JarContent implements IContent
         }
     }
 
-    public synchronized Enumeration getEntries()
+    public Enumeration getEntries()
     {
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-        }
-
         // Wrap entries enumeration to filter non-matching entries.
         Enumeration e = new EntriesEnumeration(m_jarFile.entries());
 
@@ -119,14 +109,8 @@ public class JarContent implements IContent
         return (e.hasMoreElements()) ? e : null;
     }
 
-    public synchronized byte[] getEntryAsBytes(String name) throws IllegalStateException
+    public byte[] getEntryAsBytes(String name) throws IllegalStateException
     {
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-        }
-
         // Get the embedded resource.
         InputStream is = null;
         ByteArrayOutputStream baos = null;
@@ -179,15 +163,9 @@ public class JarContent implements IContent
         }
     }
 
-    public synchronized InputStream getEntryAsStream(String name)
+    public InputStream getEntryAsStream(String name)
         throws IllegalStateException, IOException
     {
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-        }
-
         // Get the embedded resource.
         InputStream is = null;
 
@@ -212,20 +190,14 @@ public class JarContent implements IContent
         return is;
     }
 
-    public synchronized IContent getEntryAsContent(String entryName)
+    public Content getEntryAsContent(String entryName)
     {
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-
-        }
-
         // If the entry name refers to the content itself, then
         // just return it immediately.
         if (entryName.equals(FelixConstants.CLASS_PATH_DOT))
         {
-            return new JarContent(m_logger, m_configMap, m_revisionLock, m_rootDir, m_file);
+            return new JarContent(m_logger, m_configMap, m_revisionLock,
+                m_rootDir, m_file, m_jarFile);
         }
 
         // Remove any leading slash.
@@ -291,25 +263,19 @@ public class JarContent implements IContent
             }
             return new JarContent(
                 m_logger, m_configMap, m_revisionLock,
-                extractJar.getParentFile(), extractJar);
+                extractJar.getParentFile(), extractJar, null);
         }
 
         // The entry could not be found, so return null.
         return null;
     }
 
-// TODO: This will need to consider security.
-    public synchronized String getEntryAsNativeLibrary(String entryName)
+// TODO: SECURITY - This will need to consider security.
+    public String getEntryAsNativeLibrary(String entryName)
     {
         // Return result.
         String result = null;
 
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            openJarFile();
-        }
-
         // Remove any leading slash.
         entryName = (entryName.startsWith("/")) ? entryName.substring(1) : entryName;
 
@@ -332,80 +298,87 @@ public class JarContent implements IContent
                 // Since native libraries cannot be shared, we must extract a
                 // separate copy per request, so use the request library counter
                 // as part of the extracted path.
+                if (m_nativeLibMap == null)
+                {
+                    m_nativeLibMap = new HashMap();
+                }
+                Integer libCount = (Integer) m_nativeLibMap.get(entryName);
+                // Either set or increment the library count.
+                libCount = (libCount == null) ? new Integer(0) : new Integer(libCount.intValue() + 1);
+                m_nativeLibMap.put(entryName, libCount);
                 File libFile = new File(
-                    libDir, Integer.toString(m_libCount) + File.separatorChar + entryName);
-                // Increment library request counter.
-                m_libCount++;
+                    libDir, libCount.toString() + File.separatorChar + entryName);
 
                 if (!BundleCache.getSecureAction().fileExists(libFile))
                 {
-                    if (!BundleCache.getSecureAction().fileExists(libFile.getParentFile()))
+                    if (!BundleCache.getSecureAction().fileExists(libFile.getParentFile())
+                        && !BundleCache.getSecureAction().mkdirs(libFile.getParentFile()))
                     {
-                        if (!BundleCache.getSecureAction().mkdirs(libFile.getParentFile()))
+                        m_logger.log(
+                            Logger.LOG_ERROR,
+                            "Unable to create library directory.");
+                    }
+                    else
+                    {
+                        InputStream is = null;
+
+                        try
+                        {
+                            is = new BufferedInputStream(
+                                m_jarFile.getInputStream(ze),
+                                BundleCache.BUFSIZE);
+                            if (is == null)
+                            {
+                                throw new IOException("No input stream: " + entryName);
+                            }
+
+                            // Create the file.
+                            BundleCache.copyStreamToFile(is, libFile);
+
+                            // Perform exec permission command on extracted library
+                            // if one is configured.
+                            String command = (String) m_configMap.get(
+                                Constants.FRAMEWORK_EXECPERMISSION);
+                            if (command != null)
+                            {
+                                Properties props = new Properties();
+                                props.setProperty("abspath", libFile.toString());
+                                command = Util.substVars(command, "command", null, props);
+                                Process p = BundleCache.getSecureAction().exec(command);
+                                // We have to make sure we read stdout and stderr because
+                                // otherwise we will block on certain unbuffered os's
+                                // (like eg. windows)
+                                Thread stdOut = new Thread(
+                                    new DevNullRunnable(p.getInputStream()));
+                                Thread stdErr = new Thread(
+                                    new DevNullRunnable(p.getErrorStream()));
+                                stdOut.setDaemon(true);
+                                stdErr.setDaemon(true);
+                                stdOut.start();
+                                stdErr.start();
+                                p.waitFor();
+                                stdOut.join();
+                                stdErr.join();
+                            }
+
+                            // Return the path to the extracted native library.
+                            result = BundleCache.getSecureAction().getAbsolutePath(libFile);
+                        }
+                        catch (Exception ex)
                         {
                             m_logger.log(
                                 Logger.LOG_ERROR,
-                                "Unable to create library directory.");
+                                "Extracting native library.", ex);
                         }
-                        else
+                        finally
                         {
-                            InputStream is = null;
-
                             try
                             {
-                                is = new BufferedInputStream(
-                                    m_jarFile.getInputStream(ze),
-                                    BundleCache.BUFSIZE);
-                                if (is == null)
-                                {
-                                    throw new IOException("No input stream: " + entryName);
-                                }
-
-                                // Create the file.
-                                BundleCache.copyStreamToFile(is, libFile);
-
-                                // Perform exec permission command on extracted library
-                                // if one is configured.
-                                String command = (String) m_configMap.get(
-                                    Constants.FRAMEWORK_EXECPERMISSION);
-                                if (command != null)
-                                {
-                                    Properties props = new Properties();
-                                    props.setProperty("abspath", libFile.toString());
-                                    command = Util.substVars(command, "command", null, props);
-                                    Process p = BundleCache.getSecureAction().exec(command);
-                                    // We have to make sure we read stdout and stderr because otherwise
-                                    // we will block on certain unbuffered os's (like eg. windows)
-                                    Thread stdOut = new Thread(new DevNullRunnable(p.getInputStream()));
-                                    Thread stdErr = new Thread(new DevNullRunnable(p.getErrorStream()));
-                                    stdOut.setDaemon(true);
-                                    stdErr.setDaemon(true);
-                                    stdOut.start();
-                                    stdErr.start();
-                                    p.waitFor();
-                                    stdOut.join();
-                                    stdErr.join();
-                                }
-
-                                // Return the path to the extracted native library.
-                                result = BundleCache.getSecureAction().getAbsolutePath(libFile);
+                                if (is != null) is.close();
                             }
-                            catch (Exception ex)
+                            catch (IOException ex)
                             {
-                                m_logger.log(
-                                    Logger.LOG_ERROR,
-                                    "Extracting native library.", ex);
-                            }
-                            finally
-                            {
-                                try
-                                {
-                                    if (is != null) is.close();
-                                }
-                                catch (IOException ex)
-                                {
-                                    // Not much we can do.
-                                }
+                                // Not much we can do.
                             }
                         }
                     }
@@ -426,26 +399,11 @@ public class JarContent implements IContent
         return "JAR " + m_file.getPath();
     }
 
-    public synchronized File getFile()
+    public File getFile()
     {
         return m_file;
     }
 
-    private void openJarFile() throws RuntimeException
-    {
-        if (m_jarFile == null)
-        {
-            try
-            {
-                m_jarFile = BundleCache.getSecureAction().openJAR(m_file, false);
-            }
-            catch (IOException ex)
-            {
-                throw new RuntimeException("Unable to open JAR file, probably deleted: " + ex.getMessage());
-            }
-        }
-    }
-
     /**
      * This method extracts an embedded JAR file from the bundle's
      * JAR file.
@@ -509,9 +467,22 @@ public class JarContent implements IContent
         }
     }
 
+    private static JarFileX openJarFile(File file) throws RuntimeException
+    {
+        try
+        {
+            return BundleCache.getSecureAction().openJAR(file, false);
+        }
+        catch (IOException ex)
+        {
+            throw new RuntimeException(
+                "Unable to open JAR file, probably deleted: " + ex.getMessage());
+        }
+    }
+
     private static class EntriesEnumeration implements Enumeration
     {
-        private Enumeration m_enumeration = null;
+        private final Enumeration m_enumeration;
 
         public EntriesEnumeration(Enumeration enumeration)
         {
